@@ -24,6 +24,98 @@
 #include "pyorbit-private.h"
 #include <orbit/poa/poa-types.h>
 
+
+static PyObject *
+pyorbit_object_adaptor_set_thread_hint (PyCORBA_Object *self, PyObject *args)
+{
+    ORBitThreadHint hint;
+
+    if (!PyArg_ParseTuple(args, "i:ObjectAdaptor.set_thread_hint", &hint))
+	return NULL;
+    ORBit_ObjectAdaptor_set_thread_hint((ORBit_ObjectAdaptor)self->objref, hint);
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
+static PyObject *
+pyorbit_object_adaptor_get_thread_hint(PyCORBA_Object *self)
+{
+    ORBitThreadHint hint;
+
+    hint = ORBit_ObjectAdaptor_get_thread_hint((ORBit_ObjectAdaptor)self->objref);
+    return PyInt_FromLong(hint);
+}
+
+static PyObject *
+pyorbit_object_adaptor_object_bind_to_current_thread(PyObject *self)
+{
+    if (!PyObject_TypeCheck(self, &PyCORBA_Object_Type)) {
+        PyErr_SetString(PyExc_TypeError,
+                        "object_bind_to_current_thread: argument must be a corba object");
+        return NULL;
+    }
+    ORBit_ObjectAdaptor_object_bind_to_current_thread(((PyCORBA_Object *) self)->objref);
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
+static PyMethodDef pyorbit_object_adaptor_methods[] = {
+    { "set_thread_hint", (PyCFunction)pyorbit_object_adaptor_set_thread_hint,
+      METH_VARARGS, NULL },
+    { "get_thread_hint", (PyCFunction)pyorbit_object_adaptor_get_thread_hint,
+      METH_NOARGS, NULL },
+    { "object_bind_to_current_thread",
+      (PyCFunction)pyorbit_object_adaptor_object_bind_to_current_thread,
+      METH_O|METH_STATIC, NULL },
+    { NULL, NULL, 0, NULL }
+};
+
+PyTypeObject PyORBit_ObjectAdaptor_Type = {
+    PyObject_HEAD_INIT(NULL)
+    0,                                  /* ob_size */
+    "ORBit.ObjectAdaptor",              /* tp_name */
+    sizeof(PyCORBA_Object),             /* tp_basicsize */
+    0,                                  /* tp_itemsize */
+    /* methods */
+    (destructor)0,    			/* tp_dealloc */
+    (printfunc)0,                       /* tp_print */
+    (getattrfunc)0,                     /* tp_getattr */
+    (setattrfunc)0,                     /* tp_setattr */
+    (cmpfunc)0,                         /* tp_compare */
+    (reprfunc)0,                        /* tp_repr */
+    0,                                  /* tp_as_number */
+    0,                                  /* tp_as_sequence */
+    0,                                  /* tp_as_mapping */
+    (hashfunc)0,                        /* tp_hash */
+    (ternaryfunc)0,                     /* tp_call */
+    (reprfunc)0,                        /* tp_str */
+    (getattrofunc)0,                    /* tp_getattro */
+    (setattrofunc)0,                    /* tp_setattro */
+    0,                                  /* tp_as_buffer */
+    Py_TPFLAGS_DEFAULT,                 /* tp_flags */
+    NULL, /* Documentation string */
+    (traverseproc)0,                    /* tp_traverse */
+    (inquiry)0,                         /* tp_clear */
+    (richcmpfunc)0,                     /* tp_richcompare */
+    0,                                  /* tp_weaklistoffset */
+    (getiterfunc)0,                     /* tp_iter */
+    (iternextfunc)0,                    /* tp_iternext */
+    pyorbit_object_adaptor_methods,     /* tp_methods */
+    0,                                  /* tp_members */
+    0,                			/* tp_getset */
+    (PyTypeObject *)0,                  /* tp_base */
+    (PyObject *)0,                      /* tp_dict */
+    0,                                  /* tp_descr_get */
+    0,                                  /* tp_descr_set */
+    0,                                  /* tp_dictoffset */
+    (initproc)0,                        /* tp_init */
+    0,                                  /* tp_alloc */
+    0,                                  /* tp_new */
+    0,                                  /* tp_free */
+    (inquiry)0,                         /* tp_is_gc */
+    (PyObject *)0,                      /* tp_bases */
+};
+
 static void
 pyorbit_poa_dealloc(PyCORBA_Object *self)
 {
@@ -58,7 +150,7 @@ pyorbit_poa_activate_object(PyCORBA_Object *self, PyObject *args)
     if (pyorbit_check_ex(&ev))
 	return NULL;
 
-    ret = PyString_FromStringAndSize(id->_buffer, id->_length);
+    ret = PyString_FromStringAndSize((char *) id->_buffer, id->_length);
     CORBA_free(id);
     return ret;
 }
@@ -139,7 +231,7 @@ pyorbit_poa_servant_to_id(PyCORBA_Object *self, PyObject *args)
     if (pyorbit_check_ex(&ev))
 	return NULL;
 
-    ret = PyString_FromStringAndSize(id->_buffer, id->_length);
+    ret = PyString_FromStringAndSize((char *) id->_buffer, id->_length);
     CORBA_free(id);
     return ret;
 }
@@ -187,7 +279,7 @@ pyorbit_poa_reference_to_id(PyCORBA_Object *self, PyObject *args)
     if (pyorbit_check_ex(&ev))
 	return NULL;
 
-    ret = PyString_FromStringAndSize(id->_buffer, id->_length);
+    ret = PyString_FromStringAndSize((char *) id->_buffer, id->_length);
     CORBA_free(id);
     return ret;
 }
@@ -208,6 +300,76 @@ pyorbit_poa_get_the_poamanager(PyCORBA_Object *self)
     return ret;
 }
 
+static PyObject *
+pyorbit_poa_create_thread_policy(PyCORBA_Object *self, PyObject *args)
+{
+    CORBA_Environment ev;
+    PyObject *py_policy;
+    PortableServer_ThreadPolicyValue value;
+    PortableServer_ThreadPolicy policy;
+
+    if (!PyArg_ParseTuple(args, "i:POA.create_thread_policy", &value))
+	return NULL;
+    CORBA_exception_init(&ev);
+    policy = PortableServer_POA_create_thread_policy((PortableServer_POA)self->objref, value, &ev);
+    if (pyorbit_check_ex(&ev))
+	return NULL;
+    py_policy = pycorba_policy_new((CORBA_Object) policy);
+    CORBA_Object_release((CORBA_Object) policy, &ev);
+    return py_policy;
+}
+
+  /* 
+        POA create_POA(         in string       adapter_name,
+                                in POAManager   a_POAManager,
+                                in CORBA::PolicyList policies)
+                                    raises (AdapterAlreadyExists, InvalidPolicy);
+ */
+static PyObject *
+pyorbit_poa_create_POA(PyCORBA_Object *self, PyObject *args, PyObject *kwargs)
+{
+    static char *kwlist[] = { "adapter_name", "a_POAManager", "policies", NULL};
+    CORBA_Environment ev;
+    CORBA_PolicyList policies;
+    char *adapter_name;
+    PyObject *py_POAManager, *py_policies, *py_new_poa;
+    PortableServer_POA new_poa;
+    PortableServer_POAManager manager;
+    int i;
+
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "sO!O!:CORBA.ORB_init", kwlist,
+                                     &adapter_name,
+                                     &PyPortableServer_POAManager_Type, &py_POAManager,
+                                     &PyList_Type, &py_policies))
+	return NULL;
+
+    policies._length = PyList_Size(py_policies);
+    policies._maximum = policies._length;
+    policies._buffer = CORBA_PolicyList_allocbuf(policies._length);
+    for (i = 0; i < policies._length; ++i) {
+        PyObject *item = PyList_GET_ITEM(py_policies, i);
+        if (!PyObject_TypeCheck(item, &PyCORBA_Policy_Type)) {
+            CORBA_free(policies._buffer);
+            PyErr_SetString(PyExc_TypeError, "policies must be a list of CORBA.Policy objects");
+            return NULL;
+        }
+        policies._buffer[i] = (CORBA_Policy) ((PyCORBA_Policy *) item)->objref;
+    }
+
+    manager = (PortableServer_POAManager) ((PyCORBA_Object *) py_POAManager)->objref;
+    
+    CORBA_exception_init(&ev);
+    new_poa = PortableServer_POA_create_POA((PortableServer_POA)self->objref,
+                                            adapter_name, manager, &policies, &ev);
+    CORBA_free(policies._buffer);
+    if (pyorbit_check_ex(&ev))
+	return NULL;
+    py_new_poa = pyorbit_poa_new(new_poa);
+    CORBA_Object_release((CORBA_Object) new_poa, &ev);
+    return py_new_poa;
+}
+
+
 static PyMethodDef pyorbit_poa_methods[] = {
     { "activate_object", (PyCFunction)pyorbit_poa_activate_object,
       METH_VARARGS },
@@ -219,16 +381,21 @@ static PyMethodDef pyorbit_poa_methods[] = {
     { "servant_to_id", (PyCFunction)pyorbit_poa_servant_to_id, METH_VARARGS },
     { "servant_to_reference", (PyCFunction)pyorbit_poa_servant_to_reference,
       METH_VARARGS },
-    // { "reference_to_servant", (PyCFunction)pyorbit_poa_reference_to_servant,
-    //   METH_VARARGS },
+    /* { "reference_to_servant", (PyCFunction)pyorbit_poa_reference_to_servant,
+       METH_VARARGS }, */
     { "reference_to_id", (PyCFunction)pyorbit_poa_reference_to_id,
       METH_VARARGS },
-    // { "id_to_servant", (PyCFunction)pyorbit_poa_id_to_servant,
-    //   METH_VARARGS },
-    // { "id_to_reference", (PyCFunction)pyorbit_poa_id_to_reference,
-    //   METH_VARARGS },
+    /* { "id_to_servant", (PyCFunction)pyorbit_poa_id_to_servant,
+       METH_VARARGS },
+     { "id_to_reference", (PyCFunction)pyorbit_poa_id_to_reference,
+       METH_VARARGS }, */
     { "_get_the_POAManager", (PyCFunction)pyorbit_poa_get_the_poamanager,
       METH_NOARGS },
+    { "create_thread_policy", (PyCFunction)pyorbit_poa_create_thread_policy,
+      METH_VARARGS },
+    { "create_POA", (PyCFunction)pyorbit_poa_create_POA,
+      METH_KEYWORDS },
+
     { NULL, NULL, 0 }
 };
 
